# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import hashlib
import multiprocessing
import os
import re
import shutil
import time
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from eodag.api.product.metadata_mapping import ONLINE_STATUS
from tests import TEST_RESOURCES_PATH
from tests.context import (
    GENERIC_PRODUCT_TYPE,
    AuthenticationError,
    EODataAccessGateway,
    SearchResult,
    sanitize,
    uri_to_path,
)

PEPS_SEARCH_ARGS = [
    "peps",
    "S2_MSI_L1C",
    "2020-08-08",
    "2020-08-16",
    [137.772897, 13.134202, 153.749135, 23.885986],
]
GEODES_SEARCH_ARGS = [
    "geodes",
    "S2_MSI_L1C",
    "2024-08-08",
    "2024-08-16",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
AWSEOS_SEARCH_ARGS = [
    "aws_eos",
    "S2_MSI_L1C",
    "2020-01-01",
    "2020-01-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
EARTH_SEARCH_SEARCH_ARGS = [
    "earth_search",
    "S2_MSI_L1C",
    "2020-01-01",
    "2020-01-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
EARTH_SEARCH_GCS_SEARCH_ARGS = [
    "earth_search_gcs",
    "S2_MSI_L1C",
    "2020-01-01",
    "2020-01-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
USGS_SATAPI_AWS_SEARCH_ARGS = [
    "usgs_satapi_aws",
    "LANDSAT_C2L1",
    "2020-01-01",
    "2020-01-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
CREODIAS_SEARCH_ARGS = [
    "creodias",
    "S2_MSI_L1C",
    "2019-03-01",
    "2019-03-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
COP_DATASPACE_SEARCH_ARGS = [
    "cop_dataspace",
    "S2_MSI_L1C",
    "2019-03-01",
    "2019-03-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
PLANETARY_COMPUTER_SEARCH_ARGS = [
    "planetary_computer",
    "S2_MSI_L2A",
    "2019-03-01",
    "2019-03-15",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
HYDROWBEB_NEXT_SEARCH_ARGS = [
    "hydroweb_next",
    "SWOT_PRIOR_RIVER_DATABASE",
    "2022-04-01",
    "2022-04-10",
    [0.2563590566012408, 43.19555008715042, 2.379835675499976, 43.907759172380565],
]
# As of 2021-01-14 the products previously required in 2020-08 were offline.
# Trying here to retrieve the most recent products which are more likely to be online.
today = datetime.date.today()
week_span = datetime.timedelta(days=7)
day_span = datetime.timedelta(days=1)
USGS_RECENT_SEARCH_ARGS = [
    "usgs",
    "LANDSAT_C2L1",
    (today - 6 * week_span).isoformat(),
    (today - 5 * week_span).isoformat(),
    [50, 50, 50.3, 50.3],
]
USGS_OLD_SEARCH_ARGS = [
    "usgs",
    "LANDSAT_C2L1",
    "2017-03-01",
    "2017-03-10",
    [50, 50, 50.3, 50.3],
]
ECMWF_SEARCH_ARGS = [
    "ecmwf",
    "TIGGE_CF_SFC",
    "2017-03-01",
    "2017-03-02",
    # no need of an additional post-processing area extraction
    [-180, -90, 180, 90],
]
ECMWF_SEARCH_KWARGS = {
    # request for only 1 parameter instead of all available
    "param": "tcc",
}
COP_ADS_SEARCH_ARGS = [
    "cop_ads",
    "CAMS_EAC4",
    "2021-01-01",
    "2021-01-05",
    # no need of an additional post-processing area extraction
    [-180, -90, 180, 90],
]
COP_ADS_SEARCH_KWARGS = {
    # request for grib file instead of netcdf
    "format": "grib",
}
COP_CDS_SEARCH_ARGS = [
    "cop_cds",
    "ERA5_SL",
    "2021-01-01",
    "2021-01-02",
    # no need of an additional post-processing area extraction
    [-180, -90, 180, 90],
]
COP_CDS_SEARCH_KWARGS = {
    # request for grib file instead of netcdf
    "format": "grib",
}
COP_EWDS_SEARCH_ARGS = [
    "cop_ewds",
    "EFAS_SEASONAL",
    "2021-01-01",
    "2021-01-05",
    [-180, -90, 180, 90],
]
SARA_SEARCH_ARGS = [
    "sara",
    "S2_MSI_L1C",
    "2020-03-01",
    "2020-03-15",
    [150, -33, 151, -32],
]
METEOBLUE_SEARCH_ARGS = [
    "meteoblue",
    "NEMSGLOBAL_TCDC",
    (today + day_span).isoformat(),
    (today + 2 * day_span).isoformat(),
    [0.2, 43.2, 0.5, 43.5],
]
WEKEO_MAIN_SEARCH_ARGS = [
    "wekeo_main",
    "S5P_L1B_IR_ALL",
    "2018-04-30",
    "2018-11-18",
    [-180, -90, 180, 90],
]
WEKEO_ECMWF_SEARCH_ARGS = [
    "wekeo_ecmwf",
    "ERA5_SL_MONTHLY",
    "2024-01-01",
    "2024-01-01",
    [-180, -90, 180, 90],
]
EUMETSAT_DS_SEARCH_ARGS = [
    "eumetsat_ds",
    "S3_OLCI_L2WFR",
    "2021-03-11",
    "2021-03-11",
    [-69.3363, -75.9038, -39.3465, -68.2361],
]

FEDEO_CEDA_SEARCH_ARGS = [
    "fedeo_ceda",
    "MERIS_ALAMO_L2_V2.2",
    None,
    None,
    None,
]


@pytest.mark.enable_socket
class EndToEndBase(unittest.TestCase):
    def execute_search(
        self,
        provider,
        product_type,
        start,
        end,
        geom,
        offline=False,
        page=None,
        items_per_page=None,
        check_product=True,
        search_kwargs_dict={},
    ):
        """Search products on provider:

        - First set the preferred provider as the one given in parameter
        - Then do the search
        - Then ensure that at least the first result originates from the provider
        - Return one product to be downloaded
        """
        search_criteria = {
            "productType": product_type,
            "start": start,
            "end": end,
            "geom": geom,
            "raise_errors": True,
            **search_kwargs_dict,
        }
        if items_per_page:
            search_criteria["items_per_page"] = items_per_page
        if page:
            search_criteria["page"] = page
        results = self.eodag.search(provider=provider, **search_criteria)
        if offline:
            results = [
                prod
                for prod in results
                if prod.properties.get("storageStatus", "") != ONLINE_STATUS
            ]
        if check_product:
            self.assertGreater(len(results), 0)
            one_product = results[0]
            self.assertEqual(one_product.provider, provider)
            return one_product
        else:
            return results

    def execute_search_all(
        self,
        provider,
        product_type,
        start,
        end,
        geom,
        items_per_page=None,
        check_products=True,
    ):
        """Search all the products on provider:

        - First set the preferred provider as the one given in parameter
        - Then do the search_all
        - Then ensure that at least the first result originates from the provider
        - Return all the products
        """
        search_criteria = {
            "start": start,
            "end": end,
            "geom": geom,
        }
        self.eodag.set_preferred_provider(provider)
        results = self.eodag.search_all(
            productType=product_type, items_per_page=items_per_page, **search_criteria
        )
        if check_products:
            self.assertGreater(len(results), 0)
            one_product = results[0]
            self.assertEqual(one_product.provider, provider)
        return results


# @unittest.skip("skip auto run")
class TestEODagEndToEnd(EndToEndBase):
    """Make real case tests. This assume the existence of a user conf file in resources folder named user_conf.yml"""  # noqa

    @classmethod
    def setUpClass(cls):

        # use tests/resources/user_conf.yml if exists else default file ~/.config/eodag/eodag.yml
        tests_user_conf = os.path.join(TEST_RESOURCES_PATH, "user_conf.yml")
        if os.path.isfile(tests_user_conf):
            cls.eodag = EODataAccessGateway(
                user_conf_file_path=os.path.join(TEST_RESOURCES_PATH, "user_conf.yml")
            )
        else:
            cls.eodag = EODataAccessGateway()

        # temp download directory
        cls.tmp_download_dir = TemporaryDirectory()
        cls.tmp_download_path = cls.tmp_download_dir.name

        for provider, conf in cls.eodag.providers_config.items():
            # Change download directory to cls.tmp_download_path for tests
            if hasattr(conf, "download") and hasattr(conf.download, "output_dir"):
                conf.download.output_dir = cls.tmp_download_path
            elif hasattr(conf, "api") and hasattr(conf.api, "output_dir"):
                conf.api.output_dir = cls.tmp_download_path
            else:
                # no output_dir found for provider
                pass
            # Force all providers implementing RestoSearch and defining how to retrieve
            # products by specifying the
            # location scheme to use https, enabling actual downloading of the product
            if (
                getattr(getattr(conf, "search", {}), "product_location_scheme", "https")
                == "file"
            ):
                conf.search.product_location_scheme = "https"
            # Disable extraction
            try:  # Case HTTPDownload plugin
                conf.download.extract = False
            except (KeyError, AttributeError):  # case api plugin
                conf.api.extract = False

    def setUp(self):
        self.downloaded_file_path = ""

    @classmethod
    def tearDownClass(cls):
        cls.tmp_download_dir.cleanup()

    def execute_download(self, product, expected_filename, wait_sec=5, timeout_sec=120):
        """Download the product in a child process, avoiding to perform the entire
        download, then do some checks and delete the downloaded result from the
        filesystem.
        """

        start_time = time.time()

        dl_pool = multiprocessing.Pool()

        dl_result = dl_pool.apply_async(
            func=self.eodag.download,
            args=(product, None, wait_sec / 60, timeout_sec / 60),
        )
        max_wait_time = timeout_sec
        while (
            sum(
                f.stat().st_size
                for f in Path(self.tmp_download_path).glob("**/[!quicklooks]*")
                if f.is_file()
            )
            <= 0
            and max_wait_time > 0
        ):
            # check every 2s if download has start
            dl_result.wait(2)
            max_wait_time -= 2

        try:
            dl_result.get(timeout=wait_sec)
        except multiprocessing.TimeoutError:
            pass

        dl_pool.terminate()
        dl_pool.close()

        stop_time = time.time()
        print(stop_time - start_time)

        self.assertIn(
            expected_filename, os.listdir(product.downloader.config.output_dir)
        )
        self.downloaded_file_path = os.path.join(
            product.downloader.config.output_dir, expected_filename
        )
        # check whether expected_filename refers to a file or a dir
        if os.path.isdir(self.downloaded_file_path):
            product_directory = Path(self.downloaded_file_path)
            downloaded_size = sum(
                f.stat().st_size for f in product_directory.glob("**/*") if f.is_file()
            )
        else:
            downloaded_size = os.stat(self.downloaded_file_path).st_size
        # The partially downloaded file should be greater or equal to 5 KB
        self.assertGreaterEqual(downloaded_size, 5 * 2**10)

    def test_end_to_end_search_download_usgs_recent(self):
        product = self.execute_search(*USGS_RECENT_SEARCH_ARGS)
        expected_filename = "{}.tar.gz".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_usgs_old(self):
        product = self.execute_search(*USGS_OLD_SEARCH_ARGS)
        expected_filename = "{}.tar.gz".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    # @unittest.skip("service unavailable for the moment")
    def test_end_to_end_search_download_peps(self):
        product = self.execute_search(*PEPS_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_geodes(self):
        product = self.execute_search(*GEODES_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_creodias(self):
        product = self.execute_search(*CREODIAS_SEARCH_ARGS)
        self.eodag.providers_config["creodias"].auth.credentials[
            "totp"
        ] = "PLEASE_CHANGE_ME"
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_cop_dataspace(self):
        product = self.execute_search(*COP_DATASPACE_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_planetary_computer(self):
        product = self.execute_search(*PLANETARY_COMPUTER_SEARCH_ARGS)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename, wait_sec=20)

    def test_end_to_end_search_download_hydroweb_next(self):
        product = self.execute_search(*HYDROWBEB_NEXT_SEARCH_ARGS)
        expected_filename = "{}".format(
            sanitize(product.properties["title"])
            + "-"
            + sanitize(product.properties["id"])
        )
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_fedeo_ceda(self):
        self.eodag.discover_product_types(provider="fedeo_ceda")
        products = self.execute_search(*FEDEO_CEDA_SEARCH_ARGS, check_product=False)
        expected_filename = "{}".format(products[2].properties["title"])
        self.execute_download(products[2], expected_filename)

    def test_end_to_end_search_download_creodias_noresult(self):
        """Requesting a page on creodias with no results must return an empty SearchResult"""
        # As of 2021-03-19 this search at page 1 returns 31 products, so at page 2 there
        # are no products available and creodias returns a response without products (`hits`).
        product = self.execute_search(
            *CREODIAS_SEARCH_ARGS, page=2, items_per_page=50, check_product=False
        )
        self.assertEqual(len(product), 0)

    # @unittest.skip("expired aws_eos api key")
    def test_end_to_end_search_download_aws_eos(self):
        product = self.execute_search(*AWSEOS_SEARCH_ARGS)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_earth_search(self):
        product = self.execute_search(*EARTH_SEARCH_SEARCH_ARGS)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename, wait_sec=15)

    def test_end_to_end_search_download_earth_search_gcs(self):
        product = self.execute_search(*EARTH_SEARCH_GCS_SEARCH_ARGS)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename, wait_sec=20)

    def test_end_to_end_search_download_usgs_satapi_aws(self):
        product = self.execute_search(*USGS_SATAPI_AWS_SEARCH_ARGS)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename, wait_sec=15)

    def test_end_to_end_search_download_ecmwf(self):
        product = self.execute_search(
            *ECMWF_SEARCH_ARGS, search_kwargs_dict=ECMWF_SEARCH_KWARGS
        )
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_cop_ads(self):
        product = self.execute_search(
            *COP_ADS_SEARCH_ARGS, search_kwargs_dict=COP_ADS_SEARCH_KWARGS
        )
        expected_filename = "{}.grib".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_cop_cds(self):
        product = self.execute_search(
            *COP_CDS_SEARCH_ARGS, search_kwargs_dict=COP_CDS_SEARCH_KWARGS
        )
        expected_filename = "{}.grib".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_cop_ewds(self):
        product = self.execute_search(*COP_EWDS_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_sara(self):
        product = self.execute_search(*SARA_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_meteoblue(self):
        product = self.execute_search(*METEOBLUE_SEARCH_ARGS)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    def test_end_to_end_search_download_wekeo_main(self):
        product = self.execute_search(*WEKEO_MAIN_SEARCH_ARGS, page=2, items_per_page=4)
        expected_filename = "{}".format(product.properties["title"])
        self.execute_download(product, expected_filename, timeout_sec=40)

    def test_end_to_end_search_download_wekeo_ecmwf(self):
        product = self.execute_search(*WEKEO_ECMWF_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename, timeout_sec=40)

    def test_end_to_end_search_download_eumetsat_ds(self):
        product = self.execute_search(*EUMETSAT_DS_SEARCH_ARGS)
        expected_filename = "{}.zip".format(product.properties["title"])
        self.execute_download(product, expected_filename)

    # @unittest.skip("service unavailable for the moment")
    def test_get_quicklook_peps(self):
        product = self.execute_search(
            "peps", "S2_MSI_L1C", "2019-03-01", "2019-03-15", [50, 50, 50.3, 50.3]
        )
        quicklook_file_path = product.get_quicklook(filename="peps_quicklook")
        # TearDown will remove quicklook_file_path on end
        self.downloaded_file_path = quicklook_file_path

        self.assertNotEqual(quicklook_file_path, "")
        self.assertEqual(os.path.basename(quicklook_file_path), "peps_quicklook")
        self.assertEqual(
            os.path.dirname(quicklook_file_path),
            os.path.join(product.downloader.config.output_dir, "quicklooks"),
        )
        self.assertGreaterEqual(os.stat(quicklook_file_path).st_size, 2**5)

    def test__search_by_id_creodias(self):
        # A single test with creodias to check that _search_by_id returns
        # correctly the exact product looked for.
        uid = "S2A_MSIL1C_20200810T030551_N0209_R075_T53WPU_20200810T050611"
        provider = "creodias"

        products = self.eodag._search_by_id(
            uid=uid, provider=provider, productType="S2_MSI_L1C"
        )
        product = products[0]

        self.assertEqual(product.properties["id"], uid)
        self.assertIsNotNone(product.product_type)

    def test_search_by_tile(self):
        """Search by tileIdentifier should find results and correctly map found metadata"""
        # providers supporting search-by-tile
        supported_providers_product_types = [
            ("peps", "S2_MSI_L1C"),
            ("planetary_computer", "S2_MSI_L2A"),
            ("earth_search", "S2_MSI_L1C"),
        ]

        tile_id = "31TCJ"

        for provider, product_type in supported_providers_product_types:
            products = self.eodag.search(
                productType=product_type,
                start="2021-06-01",
                end="2021-06-30",
                tileIdentifier=tile_id,
                provider=provider,
            )
            self.assertGreater(len(products), 0, msg=f"no result found for {provider}")
            self.assertEqual(
                products[0].properties["tileIdentifier"],
                tile_id,
                msg=f"tileIdentifier not mapped for {provider}",
            )

    def test_end_to_end_discover_product_types_creodias(self):
        """discover_product_types() must return an external product types configuration for creodias"""
        provider = "creodias"
        ext_product_types_conf = self.eodag.discover_product_types(provider=provider)
        self.assertEqual(
            "SENTINEL-1",
            ext_product_types_conf[provider]["providers_config"]["SENTINEL-1"][
                "collection"
            ],
        )
        self.assertEqual(
            "SENTINEL-1",
            ext_product_types_conf[provider]["product_types_config"]["SENTINEL-1"][
                "title"
            ],
        )
        # check that all pre-configured product types are listed by provider
        provider_product_types = [
            v["collection"]
            for k, v in self.eodag.providers_config[provider].products.items()
            if k != GENERIC_PRODUCT_TYPE
        ]
        for provider_product_type in provider_product_types:
            self.assertIn(
                provider_product_type,
                ext_product_types_conf[provider]["providers_config"],
            )

    def test_end_to_end_discover_product_types_usgs_satapi_aws(self):
        """discover_product_types() must return an external product types configuration for usgs_satapi_aws"""
        provider = "usgs_satapi_aws"
        ext_product_types_conf = self.eodag.discover_product_types(provider=provider)
        self.assertEqual(
            "landsat-c2l1",
            ext_product_types_conf[provider]["providers_config"]["landsat-c2l1"][
                "productType"
            ],
        )
        self.assertEqual(
            "Landsat Collection 2 Level-1 Product",
            ext_product_types_conf[provider]["product_types_config"]["landsat-c2l1"][
                "title"
            ],
        )
        self.assertEqual(
            "1972-07-25T00:00:00.000Z",
            ext_product_types_conf[provider]["product_types_config"]["landsat-c2l1"][
                "missionStartDate"
            ],
        )
        # check that all pre-configured product types are listed by provider
        provider_product_types = [
            v["productType"]
            for k, v in self.eodag.providers_config[provider].products.items()
            if k != GENERIC_PRODUCT_TYPE
        ]
        for provider_product_type in provider_product_types:
            self.assertIn(
                provider_product_type,
                ext_product_types_conf[provider]["providers_config"],
            )

    def test_end_to_end_discover_product_types_earth_search(self):
        """discover_product_types() must return an external product types configuration for earth_search"""
        provider = "earth_search"
        ext_product_types_conf = self.eodag.discover_product_types(provider=provider)
        self.assertEqual(
            "sentinel-2-l1c",
            ext_product_types_conf[provider]["providers_config"]["sentinel-2-l1c"][
                "productType"
            ],
        )
        self.assertEqual(
            "Sentinel-2 Level-1C",
            ext_product_types_conf[provider]["product_types_config"]["sentinel-2-l1c"][
                "title"
            ],
        )
        self.assertEqual(
            "other",
            ext_product_types_conf[provider]["product_types_config"]["sentinel-2-l1c"][
                "license"
            ],
        )
        # check that all pre-configured product types are listed by provider
        provider_product_types = [
            v["productType"]
            for k, v in self.eodag.providers_config[provider].products.items()
            if k != GENERIC_PRODUCT_TYPE
        ]
        for provider_product_type in provider_product_types:
            self.assertIn(
                provider_product_type,
                ext_product_types_conf[provider]["providers_config"],
            )

    def test_end_to_end_discover_product_types_earth_search_gcs(self):
        """discover_product_types() must return None for earth_search_gcs"""
        provider = "earth_search_gcs"
        ext_product_types_conf = self.eodag.discover_product_types(provider=provider)
        self.assertIsNone(ext_product_types_conf[provider])

    def test_end_to_end_discover_product_types_fedeo_ceda(self):
        """discover_product_types() must return an external product types configuration for fedeo ceda"""
        provider = "fedeo_ceda"
        ext_product_types_conf = self.eodag.discover_product_types(provider=provider)
        self.assertEqual(
            ext_product_types_conf[provider]["product_types_config"][
                "MERIS_ALAMO_L2_V2.2"
            ]["ID"],
            "MERIS_ALAMO_L2_V2.2",
        )
        self.assertEqual(
            "MERIS",
            ext_product_types_conf[provider]["product_types_config"][
                "MERIS_ALAMO_L2_V2.2"
            ]["instrument"],
        )
        self.assertEqual(
            "other",
            ext_product_types_conf[provider]["product_types_config"][
                "MERIS_ALAMO_L2_V2.2"
            ]["license"],
        )


# @unittest.skip("skip auto run")
class TestEODagEndToEndComplete(EndToEndBase):
    """Make real and complete test cases that search for products, download them and
    extract them. There should be just a tiny number of these tests which can be quite
    long to run.

    There must be a user conf file in the test resources folder named user_conf.yml
    """

    @classmethod
    def setUpClass(cls):

        # use tests/resources/user_conf.yml if exists else default file ~/.config/eodag/eodag.yml
        tests_user_conf = os.path.join(TEST_RESOURCES_PATH, "user_conf.yml")
        if not os.path.isfile(tests_user_conf):
            unittest.SkipTest("Missing user conf file with credentials")
        cls.eodag = EODataAccessGateway(
            user_conf_file_path=os.path.join(TEST_RESOURCES_PATH, "user_conf.yml")
        )

        # temp download directory
        cls.tmp_download_dir = TemporaryDirectory()
        cls.tmp_download_path = cls.tmp_download_dir.name

        for provider, conf in cls.eodag.providers_config.items():
            # Change download directory to cls.tmp_download_path for tests
            if hasattr(conf, "download") and hasattr(conf.download, "output_dir"):
                conf.download.output_dir = cls.tmp_download_path
            elif hasattr(conf, "api") and hasattr(conf.api, "output_dir"):
                conf.api.output_dir = cls.tmp_download_path
            else:
                # no output_dir found for provider
                pass
            # Force all providers implementing RestoSearch and defining how to retrieve
            # products by specifying the
            # location scheme to use https, enabling actual downloading of the product
            if (
                getattr(getattr(conf, "search", {}), "product_location_scheme", "https")
                == "file"
            ):
                conf.search.product_location_scheme = "https"

    @classmethod
    def tearDownClass(cls):
        cls.tmp_download_dir.cleanup()

    def test_end_to_end_complete_peps(self):
        """Complete end-to-end test with PEPS for download and download_all"""
        # Search for products that are ONLINE and as small as possible
        today = datetime.date.today()
        month_span = datetime.timedelta(weeks=4)
        search_results = self.eodag.search(
            productType="S2_MSI_L1C",
            start=(today - month_span).isoformat(),
            end=today.isoformat(),
            geom={"lonmin": 1, "latmin": 42, "lonmax": 5, "latmax": 46},
            items_per_page=100,
            provider="peps",
        )
        prods_sorted_by_size = SearchResult(
            sorted(search_results, key=lambda p: p.properties["resourceSize"])
        )
        prods_online = [
            p for p in prods_sorted_by_size if p.properties["storageStatus"] == "ONLINE"
        ]
        if len(prods_online) < 2:
            unittest.skip(
                "Not enough ONLINE products found, update the search criteria."
            )

        # Retrieve one product to work with
        product = prods_online[0]

        prev_remote_location = product.remote_location
        prev_location = product.location
        # The expected product's archive filename is based on the product's title
        expected_product_name = f"{product.properties['title']}.zip"

        # Download the product, but DON'T extract it
        archive_file_path = self.eodag.download(product, extract=False)

        # The archive must have been downloaded
        self.assertTrue(os.path.isfile(archive_file_path))
        # Its name must be the "{product_title}.zip"
        self.assertIn(
            expected_product_name, os.listdir(product.downloader.config.output_dir)
        )
        # Its size should be >= 5 KB
        archive_size = os.stat(archive_file_path).st_size
        self.assertGreaterEqual(archive_size, 5 * 2**10)
        # The product remote_location should be the same
        self.assertEqual(prev_remote_location, product.remote_location)
        # However its location should have been update
        self.assertNotEqual(prev_location, product.location)
        # The location must follow the file URI scheme
        self.assertTrue(product.location.startswith("file://"))
        # That points to the downloaded archive
        self.assertEqual(uri_to_path(product.location), archive_file_path)
        # A .downloaded folder must have been created
        record_dir = os.path.join(self.tmp_download_path, ".downloaded")
        self.assertTrue(os.path.isdir(record_dir))
        # It must contain a file per product downloade, whose name is
        # the MD5 hash of the product's ``product_type`` and ``properties['id']``
        expected_hash = product.product_type + "-" + product.properties["id"]
        expected_hash = hashlib.md5(expected_hash.encode("utf-8")).hexdigest()
        record_file = os.path.join(record_dir, expected_hash)
        self.assertTrue(os.path.isfile(record_file))
        # Its content must be the product's remote location
        record_content = Path(record_file).read_text()
        self.assertEqual(record_content, product.remote_location)

        # The downloaded product should not be downloaded again if the download
        # method is executed again
        previous_archive_file_path = archive_file_path
        previous_location = product.location
        start_time = time.time()
        archive_file_path = self.eodag.download(product, extract=False)
        end_time = time.time()
        self.assertLess(end_time - start_time, 2)  # Should be really fast (< 2s)
        # The paths should be the same as before
        self.assertEqual(archive_file_path, previous_archive_file_path)
        self.assertEqual(product.location, previous_location)

        # If we emulate that the product has just been found, it should not
        # be downloaded again since the record file is still present.
        product.location = product.remote_location
        # Pretty much the same checks as with the previous step
        previous_archive_file_path = archive_file_path
        start_time = time.time()
        archive_file_path = self.eodag.download(product, extract=False)
        end_time = time.time()
        self.assertLess(end_time - start_time, 2)  # Should be really fast (< 2s)
        # The returned path should be the same as before
        self.assertEqual(archive_file_path, previous_archive_file_path)
        self.assertEqual(uri_to_path(product.location), archive_file_path)

        # Remove the archive
        os.remove(archive_file_path)

        # Now, the archive is removed but its associated record file
        # still exists. Downloading the product again should really
        # download it, if its location points to the remote location.
        # The product should be automatically extracted.
        product.location = product.remote_location
        product_dir_path = self.eodag.download(
            product, extract=True, delete_archive=False
        )

        # Its size should be >= 5 KB
        downloaded_size = sum(
            f.stat().st_size for f in Path(product_dir_path).glob("**/*") if f.is_file()
        )
        self.assertGreaterEqual(downloaded_size, 5 * 2**10)
        # The product remote_location should be the same
        self.assertEqual(prev_remote_location, product.remote_location)
        # However its location should have been update
        self.assertNotEqual(prev_location, product.location)
        # The location must follow the file URI scheme
        self.assertTrue(product.location.startswith("file://"))
        # The path must point to a directory
        self.assertTrue(os.path.isdir(product_dir_path))

        # The downloaded & extracted product should not be downloaded again if
        # the download method is executed again
        previous_product_dir_path = product_dir_path
        start_time = time.time()
        product_dir_path = self.eodag.download(product)
        end_time = time.time()
        self.assertLess(end_time - start_time, 2)  # Should be really fast (< 2s)
        # The paths should be the same as before
        self.assertEqual(product_dir_path, previous_product_dir_path)

        # Remove the archive and extracted product and reset the product's location
        os.remove(archive_file_path)
        shutil.rmtree(Path(product_dir_path).parent)
        product.location = product.remote_location

        # Now let's check download_all
        products = prods_sorted_by_size[:2]
        # Pass a copy because download_all empties the list
        archive_paths = self.eodag.download_all(products[:], extract=False)

        # The returned paths must point to the downloaded archives
        # Each product's location must be a URI path to the archive
        for product, archive_path in zip(products, archive_paths):
            self.assertTrue(os.path.isfile(archive_path))
            self.assertEqual(uri_to_path(product.location), archive_path)

        # Downloading the product again should not download them, since
        # they are all already there.
        prev_archive_paths = archive_paths
        start_time = time.time()
        archive_paths = self.eodag.download_all(products[:], extract=False)
        end_time = time.time()
        self.assertLess(end_time - start_time, 2)  # Should be really fast (< 2s)
        self.assertEqual(archive_paths, prev_archive_paths)


# @unittest.skip("skip auto run")
class TestEODagEndToEndWrongCredentials(EndToEndBase):
    """Make real case tests with wrong credentials. This assumes the existence of a
    wrong_credentials_cong.yml file in resources folder named user_conf.yml"""

    @classmethod
    def setUpClass(cls):
        tests_wrong_conf = os.path.join(
            TEST_RESOURCES_PATH, "wrong_credentials_conf.yml"
        )
        cls.eodag = EODataAccessGateway(user_conf_file_path=tests_wrong_conf)
        # backup os.environ as it will be modified by tests
        cls.eodag_env_pattern = re.compile(r"EODAG_\w+")
        cls.eodag_env_backup = {
            k: v for k, v in os.environ.items() if cls.eodag_env_pattern.match(k)
        }

    @classmethod
    def tearDownClass(cls):
        super(TestEODagEndToEndWrongCredentials, cls).tearDownClass()
        # restore os.environ
        for k, v in os.environ.items():
            if cls.eodag_env_pattern.match(k):
                os.environ.pop(k)
        os.environ.update(cls.eodag_env_backup)

    def test_end_to_end_wrong_credentials_peps(self):
        product = self.execute_search(*PEPS_SEARCH_ARGS)
        with self.assertRaises(AuthenticationError):
            self.eodag.download(product)

    def test_end_to_end_wrong_apikey_search_aws_eos(self):
        self.eodag.set_preferred_provider(AWSEOS_SEARCH_ARGS[0])
        with self.assertRaises(AuthenticationError):
            self.eodag.search(
                raise_errors=True,
                **dict(
                    zip(["productType", "start", "end", "geom"], AWSEOS_SEARCH_ARGS[1:])
                ),
            )

    # @unittest.skip("expired aws_eos api key")
    def test_end_to_end_good_apikey_wrong_credentials_aws_eos(self):
        # Setup
        # We retrieve correct credentials from the user_conf.yml file
        tests_user_conf = os.path.join(TEST_RESOURCES_PATH, "user_conf.yml")
        if not os.path.isfile(tests_user_conf):
            self.skipTest("user_conf.yml file with credentials not found.")
        # But we set the access key id and the secret to wrong values
        try:
            os.environ[
                "EODAG__AWS_EOS__AUTH__CREDENTIALS__AWS_ACCESS_KEY_ID"
            ] = "badaccessid"
            os.environ[
                "EODAG__AWS_EOS__AUTH__CREDENTIALS__AWS_SECRET_ACCESS_KEY"
            ] = "badsecret"
            os.environ["EODAG__AWS_EOS__AUTH__CREDENTIALS__AWS_PROFILE"] = "badsecret"

            eodag = EODataAccessGateway(
                user_conf_file_path=os.path.join(TEST_RESOURCES_PATH, "user_conf.yml")
            )
            eodag.set_preferred_provider(AWSEOS_SEARCH_ARGS[0])
            results = eodag.search(
                raise_errors=True,
                **dict(
                    zip(["productType", "start", "end", "geom"], AWSEOS_SEARCH_ARGS[1:])
                ),
            )
            self.assertGreater(len(results), 0)
            one_product = results[0]
            self.assertEqual(one_product.provider, AWSEOS_SEARCH_ARGS[0])
            with self.assertRaises(AuthenticationError):
                self.eodag.download(one_product)
        # Teardown
        finally:
            os.environ.pop("EODAG__AWS_EOS__AUTH__CREDENTIALS__AWS_ACCESS_KEY_ID")
            os.environ.pop("EODAG__AWS_EOS__AUTH__CREDENTIALS__AWS_SECRET_ACCESS_KEY")
            os.environ.pop("EODAG__AWS_EOS__AUTH__CREDENTIALS__AWS_PROFILE")

    def test_end_to_end_wrong_credentials_creodias(self):
        product = self.execute_search(*CREODIAS_SEARCH_ARGS)
        with self.assertRaises(AuthenticationError):
            self.eodag.download(product)

    def test_end_to_end_wrong_credentials_cop_dataspace(self):
        product = self.execute_search(*COP_DATASPACE_SEARCH_ARGS)
        with self.assertRaises(AuthenticationError):
            self.eodag.download(product)

    def test_end_to_end_wrong_credentials_search_usgs(self):
        # It should already fail while searching for the products.
        self.eodag.set_preferred_provider(USGS_RECENT_SEARCH_ARGS[0])
        with self.assertRaises(AuthenticationError):
            self.eodag.search(
                raise_errors=True,
                **dict(
                    zip(
                        ["productType", "start", "end", "geom"],
                        USGS_RECENT_SEARCH_ARGS[1:],
                    )
                ),
            )

    def test_end_to_end_wrong_credentials_search_meteoblue(self):
        # It should already fail while searching for the products.
        self.eodag.set_preferred_provider(METEOBLUE_SEARCH_ARGS[0])
        with self.assertRaises(AuthenticationError):
            self.eodag.search(
                raise_errors=True,
                **dict(
                    zip(
                        ["productType", "start", "end", "geom"],
                        METEOBLUE_SEARCH_ARGS[1:],
                    )
                ),
            )

    def test_end_to_end_wrong_credentials_search_hydroweb_next(self):
        # It should already fail while searching for the products.
        self.eodag.set_preferred_provider(HYDROWBEB_NEXT_SEARCH_ARGS[0])
        with self.assertRaises(AuthenticationError):
            self.eodag.search(
                raise_errors=True,
                **dict(
                    zip(
                        ["productType", "start", "end", "geom"],
                        HYDROWBEB_NEXT_SEARCH_ARGS[1:],
                    )
                ),
            )

    def test_end_to_end_wrong_credentials_search_wekeo_main(self):
        # It should already fail while searching for the products.
        self.eodag.set_preferred_provider(WEKEO_MAIN_SEARCH_ARGS[0])
        with self.assertRaises(AuthenticationError):
            results, _ = self.eodag.search(
                raise_errors=True,
                **dict(
                    zip(
                        ["productType", "start", "end", "geom"],
                        WEKEO_MAIN_SEARCH_ARGS[1:],
                    )
                ),
            )

    def test_end_to_end_wrong_credentials_search_wekeo_ecmwf(self):
        # It should already fail while searching for the products.
        self.eodag.set_preferred_provider(WEKEO_ECMWF_SEARCH_ARGS[0])
        with self.assertRaises(AuthenticationError):
            self.eodag.search(
                raise_errors=True,
                **dict(
                    zip(
                        ["productType", "start", "end", "geom"],
                        WEKEO_ECMWF_SEARCH_ARGS[1:],
                    )
                ),
            )

    def test_end_to_end_wrong_credentials_search_eumetsat_ds(self):
        product = self.execute_search(*EUMETSAT_DS_SEARCH_ARGS)
        with self.assertRaises(AuthenticationError):
            self.eodag.download(product)



import unittest
from unittest import mock

from pydantic import ValidationError

from eodag.rest.types import eodag_search, stac_search


class TestStacSearch(unittest.TestCase):
    def test_sortby(self):
        # Test with valid field and direction
        sortby = stac_search.SortBy(field="test", direction="asc")
        self.assertEqual(sortby.field, "test")
        self.assertEqual(sortby.direction, "asc")

        # Test with invalid direction
        with self.assertRaises(ValidationError) as context:
            stac_search.SortBy(field="test", direction="invalid")
        self.assertTrue("Input should be 'asc' or 'desc'" in str(context.exception))

        # Test with empty field
        with self.assertRaises(ValidationError) as context:
            stac_search.SortBy(field="", direction="asc")

    def test_sortby2list(self):
        # Test with no input
        self.assertIsNone(stac_search.sortby2list(None))

        # Test with valid input
        sortby_list = stac_search.sortby2list("test,+test2,-test3")
        self.assertIsNotNone(sortby_list)
        self.assertEqual(len(sortby_list), 3)
        self.assertEqual(sortby_list[0].field, "test")
        self.assertEqual(sortby_list[0].direction, "asc")
        self.assertEqual(sortby_list[1].field, "test2")
        self.assertEqual(sortby_list[1].direction, "asc")
        self.assertEqual(sortby_list[2].field, "test3")
        self.assertEqual(sortby_list[2].direction, "desc")


class TestSearchPostRequest(unittest.TestCase):
    def test_limit(self):
        # Test with positive integer
        try:
            stac_search.SearchPostRequest.model_validate({"limit": 10})
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

        # Test with non-positive integer
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate({"limit": -10})

    def test_page(self):
        # Test with positive integer
        try:
            stac_search.SearchPostRequest.model_validate({"page": 10})
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

        # Test with non-positive integer
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate({"page": -10})

    def test_check_filter_lang(self):
        # Test with filter and without filter-lang
        values = {"filter": {"test": "value"}}
        search_args = stac_search.SearchPostRequest.model_validate(values).model_dump(
            exclude_none=True
        )
        self.assertEqual(
            search_args, {"filter_lang": "cql2-json", "filter": {"test": "value"}}
        )

        # Test with filter-lang and without filter
        values = {"filter-lang": "cql2-json"}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with incorrect filter-lang
        values = {"filter": {"test": "value"}, "filter-lang": "incorrect"}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with correct filter and filter-lang
        values = {"filter": {"test": "value"}, "filter-lang": "cql2-json"}
        try:
            stac_search.SearchPostRequest.model_validate(values)
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

    def test_str_to_str_list(self):
        # Test with string
        values = {"ids": "id1,id2", "collections": "col1,col2"}
        req = stac_search.SearchPostRequest.model_validate(values)
        self.assertEqual(req.ids, ["id1", "id2"])
        self.assertEqual(req.collections, ["col1", "col2"])

        # Test with list of strings
        values = {"ids": ["id1", "id2"], "collections": ["col1", "col2"]}
        req = stac_search.SearchPostRequest.model_validate(values)
        self.assertEqual(req.ids, ["id1", "id2"])
        self.assertEqual(req.collections, ["col1", "col2"])

    def test_validate_spatial(self):
        # Test with both bbox and intersects
        values = {
            "bbox": [-180, -90, 180, 90],
            "intersects": {"type": "Point", "coordinates": [0, 0]},
        }
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with incorrect geometry
        values = {"intersects": {"type": "Incorrect", "coordinates": [0, 0]}}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with correct geometry
        values = {"intersects": {"type": "Point", "coordinates": [0, 0]}}
        try:
            stac_search.SearchPostRequest.model_validate(values)
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

    def test_validate_bbox(self):
        # Test with incorrect bbox
        values = {"bbox": [180, -90, -180, 90]}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with correct bbox
        values = {"bbox": [-180, -90, 180, 90]}
        try:
            stac_search.SearchPostRequest.model_validate(values)
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

    def test_validate_datetime(self):
        # Test with single datetime
        values = {"datetime": "2023-12-18T19:56:32Z"}
        try:
            stac_search.SearchPostRequest.model_validate(values)
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

        # Test with datetime interval
        values = {"datetime": "2023-12-18T19:56:32Z/2023-12-19T19:56:32Z"}
        try:
            stac_search.SearchPostRequest.model_validate(values)
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

        # Test with open datetime interval
        values = {"datetime": "../2023-12-19T19:56:32Z"}
        try:
            stac_search.SearchPostRequest.model_validate(values)
        except ValidationError:
            self.fail("SearchPostRequest raised ValidationError unexpectedly!")

        # Test with both ends of range open
        values = {"datetime": "../.."}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with invalid datetime range
        values = {"datetime": "2023-12-19T19:56:32Z/2023-12-18T19:56:32Z"}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)

        # Test with invalid datetime
        values = {"datetime": "invalid"}
        with self.assertRaises(ValidationError):
            stac_search.SearchPostRequest.model_validate(values)


class TestEODAGSearch(unittest.TestCase):
    def test_remove_custom_extensions(self):
        values = {"unk:test1": "value", "oseo:test2": "value", "collections": ["value"]}
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "test1": "value",
                "test2": "value",
                "productType": "value",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_remove_keys(self):
        values = {
            "datetime": "2023-12-18T16:41:35Z",
            "collections": ["value"],
            "bbox": "value",
            "intersects": {"type": "Point", "coordinates": [1, 1]},
            "raise_errors": False,
        }
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "productType": "value",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_assemble_geom(self):
        values = {
            "geometry": {"type": "Point", "coordinates": [1, 1]},
            "collections": ["value"],
        }
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "geom": {"type": "Point", "coordinates": [1, 1]},
                "productType": "value",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_convert_collections_to_product_type(self):
        # Test with one collection in collections
        values = {"collections": ["test_collections"], "test": "value"}
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "test": "value",
                "productType": "test_collections",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

        # Test with one collection in collection
        values = {"collections": ["test_collection"], "test": "value"}
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "test": "value",
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

        # Test with more than one collection in collections
        values = {"collections": ["value1", "value2"], "test": "value"}
        with self.assertRaises(ValidationError) as context:
            eodag_search.EODAGSearch.model_validate(values)
        self.assertTrue(
            "Only one collection is supported per search" in str(context.exception)
        )

    def test_convert_query_to_dict(self):
        # Test with valid query
        values = {
            "query": {"properties.test": {"eq": "value"}},
            "collections": ["test_collection"],
        }
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "test": "value",
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

        # Test with invalid query syntax
        values = {"query": {"invalid": "invalid"}, "collections": ["test_collection"]}
        with self.assertRaises(ValidationError) as context:
            eodag_search.EODAGSearch.model_validate(values)
        self.assertTrue(
            "Exactly 1 operator must be specified per property"
            in str(context.exception)
        )

        # Test with multiple operators for a property
        values = {
            "query": {"properties.test": {"eq": "value", "lte": "value"}},
            "collections": ["test_collection"],
        }
        with self.assertRaises(ValidationError) as context:
            eodag_search.EODAGSearch.model_validate(values)
        self.assertTrue(
            "Exactly 1 operator must be specified per property"
            in str(context.exception)
        )

        # Test with unsupported operator
        values = {
            "query": {"properties.test": {"neq": "value"}},
            "collections": ["test_collection"],
        }
        with self.assertRaises(ValidationError) as context:
            eodag_search.EODAGSearch.model_validate(values)
        self.assertTrue(
            'operator "neq" is not supported for property "test"'
            in str(context.exception)
        )

        # Test with "lte" operator for a non-cloud_cover property
        values = {
            "query": {"properties.test": {"lte": "value"}},
            "collections": ["test_collection"],
        }
        with self.assertRaises(ValidationError) as context:
            eodag_search.EODAGSearch.model_validate(values)
        self.assertTrue(
            'operator "lte" is not supported for property "test"'
            in str(context.exception)
        )

    @mock.patch("eodag.rest.types.eodag_search.EodagEvaluator")
    @mock.patch("eodag.rest.types.eodag_search.parse_json")
    @mock.patch("eodag.rest.types.eodag_search.is_dict_str_any")
    def test_parse_cql(
        self, mock_is_dict_str_any, mock_parse_json, mock_EodagEvaluator
    ):
        # Set up the mock objects
        mock_parse_json.return_value = "parsed_filter"
        mock_EodagEvaluator.return_value.evaluate.return_value = {
            "parsed_result": "parsed_value"
        }
        mock_is_dict_str_any.return_value = True

        # Test with no filter
        values = {"test": "value"}
        expected_result = {"test": "value"}
        self.assertEqual(eodag_search.EODAGSearch.parse_cql(values), expected_result)

        # Test with valid filter
        values = {"filter": "filter", "test": "value"}
        expected_result = {"test": "value", "parsed_result": "parsed_value"}
        self.assertEqual(eodag_search.EODAGSearch.parse_cql(values), expected_result)

        # Test with invalid filter
        values = {"filter": "filter", "test": "value"}
        mock_is_dict_str_any.return_value = False
        with self.assertRaises(ValueError) as context:
            eodag_search.EODAGSearch.parse_cql(values)
        self.assertTrue(
            "The parsed filter is not a proper dictionary" in str(context.exception)
        )

    def test_join_instruments(self):
        # Test with string
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(
                {"instruments": "value", "collections": ["test_collection"]}
            ).model_dump(exclude_none=True),
            {
                "instrument": "value",
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

        # Test with list of strings
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(
                {
                    "instruments": ["value1", "value2"],
                    "collections": ["test_collection"],
                }
            ).model_dump(exclude_none=True),
            {
                "instrument": "value1,value2",
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_convert_stac_to_eodag_sortby(self):
        values = {
            "sortby": [{"field": "test", "direction": "desc"}],
            "collections": ["test_collection"],
        }
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(values).model_dump(
                exclude_none=True
            ),
            {
                "sort_by": [("test", "desc")],
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_verify_producttype_is_present(self):
        # Test with no collections and no collection
        with self.assertRaises(ValidationError) as context:
            eodag_search.EODAGSearch.model_validate({})
        self.assertTrue("A collection is required" in str(context.exception))

        # test with neither collections nor collection but isCatalog
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(
                {}, context={"isCatalog": True}
            ).model_dump(exclude_none=True),
            {
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_cleanup_dates(self):
        # Test with date ending with "+00:00"
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(
                {
                    "start": "2023-12-18T16:41:35+00:00",
                    "end": "2023-12-19T16:41:35+00:00",
                    "collections": ["test_collection"],
                }
            ).model_dump(exclude_none=True),
            {
                "start": "2023-12-18T16:41:35Z",
                "end": "2023-12-19T16:41:35Z",
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

        # Test with date not ending with "+00:00"
        self.assertEqual(
            eodag_search.EODAGSearch.model_validate(
                {
                    "start": "2023-12-20T16:41:35",
                    "end": "2023-12-21T16:41:35",
                    "collections": ["test_collection"],
                }
            ).model_dump(exclude_none=True),
            {
                "start": "2023-12-20T16:41:35",
                "end": "2023-12-21T16:41:35",
                "productType": "test_collection",
                "items_per_page": 20,
                "page": 1,
                "raise_errors": False,
            },
        )

    def test_alias_to_property(self):
        # Test with valid aliases
        self.assertEqual(
            eodag_search.EODAGSearch.to_eodag("collections"), "productType"
        )
        self.assertEqual(
            eodag_search.EODAGSearch.to_eodag("start_datetime"),
            "startTimeFromAscendingNode",
        )
        self.assertEqual(
            eodag_search.EODAGSearch.to_eodag("platform"),
            "platformSerialIdentifier",
        )

        # Test with invalid alias
        self.assertEqual(
            eodag_search.EODAGSearch.to_eodag("invalid_alias"), "invalid_alias"
        )

        # Test with empty string
        self.assertEqual(eodag_search.EODAGSearch.to_eodag(""), "")

        # Test with alias that has no corresponding property
        self.assertEqual(eodag_search.EODAGSearch.to_eodag("provider"), "provider")



# -*- coding: utf-8 -*-
# Copyright 2023, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import time
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Optional, cast

import requests

from eodag import EOProduct
from eodag.api.product.metadata_mapping import (
    format_query_params,
    mtd_cfg_as_conversion_and_querypath,
    properties_from_json,
)
from eodag.plugins.search import PreparedSearch
from eodag.plugins.search.base import Search
from eodag.types.queryables import Queryables
from eodag.utils import (
    DEFAULT_ITEMS_PER_PAGE,
    DEFAULT_MISSION_START_DATE,
    DEFAULT_PAGE,
    DEFAULT_SEARCH_TIMEOUT,
    GENERIC_PRODUCT_TYPE,
    HTTP_REQ_TIMEOUT,
    USER_AGENT,
    _deprecated,
    deepcopy,
    string_to_jsonpath,
)
from eodag.utils.exceptions import (
    NotAvailableError,
    RequestError,
    TimeOutError,
    ValidationError,
)

if TYPE_CHECKING:
    from eodag.config import PluginConfig

logger = logging.getLogger("eodag.search.data_request_search")


@_deprecated(
    reason="Plugin that was used in previous wekeo provider configuration, but not used anymore",
    version="3.7.1",
)
class DataRequestSearch(Search):
    """
    Plugin to execute search requests composed of several steps:

    #. do a data request which defines which data shall be searched
    #. check the status of the request job
    #. if finished - fetch the result of the job

    :param provider: provider name
    :param config: Search plugin configuration:

        * :attr:`~eodag.config.PluginConfig.api_endpoint` (``str``) (**mandatory**): The endpoint of the
          provider's search interface
        * :attr:`~eodag.config.PluginConfig.results_entry` (``str``) (**mandatory**): The name of
          the key in the provider search result that gives access to the result entries
        * :attr:`~eodag.config.PluginConfig.data_request_url` (``str``) (**mandatory**): url
          to which the data request shall be sent
        * :attr:`~eodag.config.PluginConfig.status_url` (``str``) (**mandatory**): url to fetch
          the status of the data request
        * :attr:`~eodag.config.PluginConfig.result_url` (``str``) (**mandatory**): url to fetch
          the search result when the data request is done
        * :attr:`~eodag.config.PluginConfig.need_auth` (``bool``): if authentication is needed for
          the search request; default: ``False``
        * :attr:`~eodag.config.PluginConfig.auth_error_code` (``int``): which error code is returned in case of an
          authentication error; only used if ``need_auth=true``
        * :attr:`~eodag.config.PluginConfig.ssl_verify` (``bool``): if the ssl certificates should be
          verified in requests; default: ``True``
        * :attr:`~eodag.config.PluginConfig.timeout` (``int``): time to wait until request timeout in seconds;
          default: ``5``
        * :attr:`~eodag.config.PluginConfig.dates_required` (``bool``): if date parameters are mandatory
          in the request; default: ``True``
        * :attr:`~eodag.config.PluginConfig.pagination` (:class:`~eodag.config.PluginConfig.Pagination`)
          (**mandatory**): The configuration of how the pagination is done on the provider. It is a tree with the
          following nodes:

          * :attr:`~eodag.config.PluginConfig.Pagination.total_items_nb_key_path` (``str``):  An XPath or JsonPath
            leading to the total number of results satisfying a request. This is used for providers which provides the
            total results metadata along with the result of the query and don't have an endpoint for querying
            the number of items satisfying a request, or for providers for which the count endpoint
            returns a json or xml document
          * :attr:`~eodag.config.PluginConfig.Pagination.max_items_per_page` (``int``): The maximum
            number of items per page that the provider can handle; default: ``50``
          * :attr:`~eodag.config.PluginConfig.Pagination.start_page` (``int``): number of the
            first page; default: ``1``

        * :attr:`~eodag.config.PluginConfig.discover_product_types`
          (:class:`~eodag.config.PluginConfig.DiscoverProductTypes`): configuration for product type discovery based on
          information from the provider; It contains the keys:

          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.fetch_url` (``str``) (**mandatory**): url from which
            the product types can be fetched
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.result_type` (``str``): type of the provider result;
            currently only ``json`` is supported (other types could be used in an extension of this plugin)
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.results_entry` (``str``) (**mandatory**): json path
            to the list of product types
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.generic_product_type_id` (``str``): mapping for the
            product type id
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.generic_product_type_parsable_metadata`
            (``dict[str, str]``): mapping for product type metadata (e.g. ``abstract``, ``licence``) which can be parsed
            from the provider result
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.generic_product_type_parsable_properties`
            (``dict[str, str]``): mapping for product type properties which can be parsed from the result and are not
            product type metadata
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.single_collection_fetch_url` (``str``): url to fetch
            data for a single collection; used if product type metadata is not available from the endpoint given in
            :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.fetch_url`
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.single_collection_fetch_qs` (``str``): query string
            to be added to the :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.fetch_url` to filter for a
            collection
          * :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.single_product_type_parsable_metadata`
            (``dict[str, str]``): mapping for product type metadata returned by the endpoint given in
            :attr:`~eodag.config.PluginConfig.DiscoverProductTypes.single_collection_fetch_url`.

        * :attr:`~eodag.config.PluginConfig.constraints_file_url` (``str``): url to fetch the constraints for a specific
          product type, can be an http url or a path to a file; the constraints are used to build queryables
        * :attr:`~eodag.config.PluginConfig.constraints_entry` (``str``): key in the json result where the constraints
          can be found; if not given, it is assumed that the constraints are on top level of the result, i.e.
          the result is an array of constraints
        * :attr:`~eodag.config.PluginConfig.metadata_mapping` (``dict[str, Any]``): The search plugins of this kind can
          detect when a metadata mapping is "query-able", and get the semantics of how to format the query string
          parameter that enables to make a query on the corresponding metadata. To make a metadata query-able,
          just configure it in the metadata mapping to be a list of 2 items, the first one being the
          specification of the query string search formatting. The later is a string following the
          specification of Python string formatting, with a special behaviour added to it. For example,
          an entry in the metadata mapping of this kind::

                completionTimeFromAscendingNode:
                    - 'f=acquisition.endViewingDate:lte:{completionTimeFromAscendingNode#timestamp}'
                    - '$.properties.acquisition.endViewingDate'

          means that the search url will have a query string parameter named ``f`` with a value of
          ``acquisition.endViewingDate:lte:1543922280.0`` if the search was done with the value
          of ``completionTimeFromAscendingNode`` being ``2018-12-04T12:18:00``. What happened is that
          ``{completionTimeFromAscendingNode#timestamp}`` was replaced with the timestamp of the value
          of ``completionTimeFromAscendingNode``. This example shows all there is to know about the
          semantics of the query string formatting introduced by this plugin: any eodag search parameter
          can be referenced in the query string with an additional optional conversion function that
          is separated from it by a ``#`` (see :func:`~eodag.api.product.metadata_mapping.format_metadata` for further
          details on the available converters). Note that for the values in the
          :attr:`~eodag.config.PluginConfig.free_text_search_operations` configuration parameter follow the same rule.
          If the metadata_mapping is not a list but only a string, this means that the parameters is not queryable but
          it is included in the result obtained from the provider. The string indicates how the provider result should
          be mapped to the eodag parameter.

    """

    data_request_id: Optional[str]

    def __init__(self, provider: str, config: PluginConfig) -> None:
        super().__init__(provider, config)
        self.config.__dict__.setdefault("result_type", "json")
        self.config.__dict__.setdefault("results_entry", "content")
        self.config.__dict__.setdefault("pagination", {})
        self.config.__dict__.setdefault("free_text_search_operations", {})
        for product_type in self.config.products.keys():
            if "metadata_mapping" in self.config.products[product_type].keys():
                self.config.products[product_type][
                    "metadata_mapping"
                ] = mtd_cfg_as_conversion_and_querypath(
                    self.config.products[product_type]["metadata_mapping"]
                )
                # Complete and ready to use product type specific metadata-mapping
                product_type_metadata_mapping = deepcopy(self.config.metadata_mapping)

                # update config using provider product type definition metadata_mapping
                # from another product
                other_product_for_mapping = self.config.products[product_type].get(
                    "metadata_mapping_from_product", ""
                )
                if other_product_for_mapping:
                    other_product_type_def_params = self.get_product_type_def_params(
                        other_product_for_mapping,
                    )
                    product_type_metadata_mapping.update(
                        other_product_type_def_params.get("metadata_mapping", {})
                    )
                # from current product
                product_type_metadata_mapping.update(
                    self.config.products[product_type]["metadata_mapping"]
                )

                self.config.products[product_type][
                    "metadata_mapping"
                ] = product_type_metadata_mapping

        if (
            self.config.result_type == "json"
            and "next_page_url_key_path" in self.config.pagination
        ):
            self.config.pagination["next_page_url_key_path"] = string_to_jsonpath(
                self.config.pagination.get("next_page_url_key_path")
            )
        self.download_info: dict[str, Any] = {}
        self.data_request_id = None

    def discover_product_types(self, **kwargs: Any) -> Optional[dict[str, Any]]:
        """Fetch product types is disabled for `DataRequestSearch`

        :returns: empty dict
        """
        return None

    def clear(self) -> None:
        """Clear search context"""
        super().clear()
        self.data_request_id = None

    def query(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> tuple[list[EOProduct], Optional[int]]:
        """
        performs the search for a provider where several steps are required to fetch the data
        """
        if kwargs.get("sort_by"):
            raise ValidationError(f"{self.provider} does not support sorting feature")

        product_type = kwargs.get("productType")

        if product_type is None:
            raise ValidationError("Required productType is missing")

        # replace "product_type" to "providerProductType" in search args if exists
        # for compatibility with DataRequestSearch method
        if kwargs.get("product_type"):
            kwargs["providerProductType"] = kwargs.pop("product_type", None)
        provider_product_type = cast(str, self._map_product_type(product_type or ""))
        keywords = {k: v for k, v in kwargs.items() if k != "auth" and v is not None}

        if provider_product_type and provider_product_type != GENERIC_PRODUCT_TYPE:
            keywords["productType"] = provider_product_type
        else:
            keywords["productType"] = product_type

        # provider product type specific conf
        self.product_type_def_params = self.get_product_type_def_params(
            product_type, format_variables=kwargs
        )

        # update config using provider product type definition metadata_mapping
        # from another product
        other_product_for_mapping = self.product_type_def_params.get(
            "metadata_mapping_from_product", ""
        )
        if other_product_for_mapping:
            other_product_type_def_params = self.get_product_type_def_params(
                other_product_for_mapping, format_variables=kwargs
            )
            self.config.metadata_mapping.update(
                other_product_type_def_params.get("metadata_mapping", {})
            )
        # from current product
        self.config.metadata_mapping.update(
            self.product_type_def_params.get("metadata_mapping", {})
        )

        # if product_type_def_params is set, remove product_type as it may conflict with this conf
        if self.product_type_def_params:
            keywords.pop("productType", None)
        keywords.update(
            {
                k: v
                for k, v in self.product_type_def_params.items()
                if k not in keywords.keys()
                and k in self.config.metadata_mapping.keys()
                and isinstance(self.config.metadata_mapping[k], list)
            }
        )

        # update dates if needed
        if getattr(self.config, "dates_required", True) and "id" not in keywords:
            if not keywords.get("startTimeFromAscendingNode"):
                keywords["startTimeFromAscendingNode"] = getattr(
                    self.config, "product_type_config", {}
                ).get("missionStartDate", DEFAULT_MISSION_START_DATE)
            if not keywords.get("completionTimeFromAscendingNode"):
                keywords["completionTimeFromAscendingNode"] = getattr(
                    self.config, "product_type_config", {}
                ).get("missionEndDate", datetime.now(timezone.utc).isoformat())

        # ask for data_request_id if not set (it must exist when iterating over pages)
        if not self.data_request_id:
            data_request_id = self._create_data_request(
                provider_product_type, product_type, **keywords
            )
            self.data_request_id = data_request_id
            request_finished = False
        else:
            data_request_id = self.data_request_id
            request_finished = True

        # loop to check search job status
        search_timeout = int(getattr(self.config, "timeout", DEFAULT_SEARCH_TIMEOUT))
        logger.info(
            f"checking status of request job {data_request_id} (timeout={search_timeout}s)"
        )
        check_beginning = datetime.now()
        while not request_finished:
            request_finished = self._check_request_status(data_request_id)
            if not request_finished and datetime.now() >= check_beginning + timedelta(
                seconds=search_timeout
            ):
                self._cancel_request(data_request_id)
                raise NotAvailableError(
                    f"Timeout reached when checking search job status for {self.provider}"
                )
            elif not request_finished:
                time.sleep(1)

        logger.info("search job for product_type %s finished", provider_product_type)
        result = self._get_result_data(
            data_request_id,
            kwargs.get("items_per_page", DEFAULT_ITEMS_PER_PAGE),
            kwargs.get("page", DEFAULT_PAGE),
        )
        # if exists, add the geometry from search args in the content of the response for each product
        if keywords.get("geometry"):
            for product_content in result["content"]:
                if product_content["extraInformation"] is None:
                    product_content["extraInformation"] = {
                        "footprint": keywords["geometry"]
                    }
                elif not product_content["extraInformation"].get("footprint"):
                    product_content["extraInformation"]["footprint"] = keywords[
                        "geometry"
                    ]
        logger.info("result retrieved from search job")
        if self._check_uses_custom_filters(product_type):
            result = self._apply_additional_filters(
                result, self.config.products[product_type]["custom_filters"]
            )
        return self._convert_result_data(
            result, data_request_id, product_type or "", **kwargs
        )

    def _create_data_request(
        self, product_type: str, eodag_product_type: str, **kwargs: Any
    ) -> str:
        headers = getattr(self.auth, "headers", USER_AGENT)
        ssl_verify = getattr(self.config.ssl_verify, "ssl_verify", True)
        try:
            url = self.config.data_request_url
            try:
                request_body = format_query_params(
                    eodag_product_type, self.config, kwargs
                )
            except ValidationError as err:
                not_queryable_search_param = Queryables.get_queryable_from_alias(
                    str(err.message).split(":")[-1].strip()
                )
                raise ValidationError(
                    f"Search parameters which are not queryable are disallowed for {product_type} on "
                    f"{self.provider}: please remove '{not_queryable_search_param}' from your search parameters",
                    {not_queryable_search_param},
                ) from err
            logger.debug(
                f"Sending search job request to {url} with {str(request_body)}"
            )
            request_job = requests.post(
                url,
                json=request_body,
                headers=headers,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
            )
            request_job.raise_for_status()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
        except requests.RequestException as e:
            raise RequestError.from_error(
                e, f"search job for product_type {product_type} could not be created"
            ) from e
        else:
            logger.info("search job for product_type %s created", product_type)
            return request_job.json()["jobId"]

    def _cancel_request(self, data_request_id: str) -> None:
        logger.info("deleting request job %s", data_request_id)
        delete_url = f"{self.config.data_request_url}/{data_request_id}"
        headers = getattr(self.auth, "headers", USER_AGENT)
        try:
            delete_resp = requests.delete(
                delete_url, headers=headers, timeout=HTTP_REQ_TIMEOUT
            )
            delete_resp.raise_for_status()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
        except requests.RequestException as e:
            raise RequestError.from_error(e, "_cancel_request failed") from e

    def _check_request_status(self, data_request_id: str) -> bool:
        logger.debug("checking status of request job %s", data_request_id)
        status_url = self.config.status_url + data_request_id
        headers = getattr(self.auth, "headers", USER_AGENT)
        ssl_verify = getattr(self.config, "ssl_verify", True)

        try:
            status_resp = requests.get(
                status_url,
                headers=headers,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
            )
            status_resp.raise_for_status()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
        except requests.RequestException as e:
            raise RequestError.from_error(e, "_check_request_status failed") from e
        else:
            status_data = status_resp.json()
            if "status_code" in status_data and status_data["status_code"] in [
                403,
                404,
            ]:
                logger.error(f"_check_request_status failed: {status_data}")
                error = RequestError("authentication token expired during request")
                error.status_code = status_data["status_code"]
                raise error
            if status_data["status"] == "failed":
                logger.error(f"_check_request_status failed: {status_data}")
                raise RequestError(
                    f"data request job has failed, message: {status_data['message']}"
                )
            return status_data["status"] == "completed"

    def _get_result_data(
        self, data_request_id: str, items_per_page: int, page: int
    ) -> dict[str, Any]:
        page = page - 1 + self.config.pagination.get("start_page", 1)
        url = self.config.result_url.format(
            jobId=data_request_id, items_per_page=items_per_page, page=page
        )
        ssl_verify = getattr(self.config, "ssl_verify", True)
        headers = getattr(self.auth, "headers", USER_AGENT)
        try:
            return requests.get(
                url, headers=headers, timeout=HTTP_REQ_TIMEOUT, verify=ssl_verify
            ).json()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
        except requests.RequestException:
            logger.error(f"Result could not be retrieved for {url}")
        return {}

    def _convert_result_data(
        self,
        result_data: dict[str, Any],
        data_request_id: str,
        product_type: str,
        **kwargs: Any,
    ) -> tuple[list[EOProduct], int]:
        """Build EOProducts from provider results"""
        results_entry = self.config.results_entry
        results = result_data[results_entry]
        logger.debug(
            "Adapting %s plugin results to eodag product representation" % len(results)
        )
        products: list[EOProduct] = []
        for result in results:
            product = EOProduct(
                self.provider,
                properties_from_json(
                    result,
                    self.get_metadata_mapping(kwargs.get("productType")),
                    discovery_config=getattr(self.config, "discover_metadata", {}),
                ),
                **kwargs,
            )
            # use product_type_config as default properties
            product.properties = dict(
                getattr(self.config, "product_type_config", {}), **product.properties
            )
            products.append(product)
        # postprocess filtering needed when provider does not natively offer filtering by id
        if "id" in kwargs:
            products = [
                p for p in products if product.properties["id"] == kwargs["id"]
            ] or products
        total_items_nb_key_path = string_to_jsonpath(
            self.config.pagination["total_items_nb_key_path"]
        )
        found_total_items_nb_paths = total_items_nb_key_path.find(results)
        if found_total_items_nb_paths and not isinstance(
            found_total_items_nb_paths, int
        ):
            total_items_nb = found_total_items_nb_paths[0].value
        else:
            total_items_nb = 0
        for p in products:
            # add the request id to the order link property (required to create data order)
            p.properties["orderLink"] = p.properties["orderLink"].replace(
                "requestJobId", str(data_request_id)
            )
            if self.config.products[product_type].get("storeDownloadUrl", False):
                # store download information to retrieve it later in case search by id
                # is not possible
                self.download_info[p.properties["id"]] = {
                    "requestJobId": data_request_id,
                    "orderLink": p.properties["orderLink"],
                    "downloadLink": p.properties["downloadLink"],
                    "provider": self.provider,
                }
        return products, total_items_nb

    def _check_uses_custom_filters(self, product_type: str) -> bool:
        if (
            product_type in self.config.products
            and "custom_filters" in self.config.products[product_type]
        ):
            return True
        return False

    def _apply_additional_filters(
        self, result: dict[str, Any], custom_filters: dict[str, str]
    ) -> dict[str, Any]:
        filtered_result = []
        results_entry = self.config.results_entry
        results = result[results_entry]
        path = string_to_jsonpath(custom_filters["filter_attribute"])
        indexes = custom_filters["indexes"].split("-")
        for record in results:
            found_paths = path.find(record)
            if not found_paths or isinstance(found_paths, int):
                continue
            filter_param = found_paths[0].value
            filter_value = filter_param[int(indexes[0]) : int(indexes[1])]
            filter_clause = "'" + filter_value + "' " + custom_filters["filter_clause"]
            if eval(filter_clause):
                filtered_result.append(record)
        result[results_entry] = filtered_result
        return result

    def _map_product_type(self, product_type: Optional[str]) -> Optional[str]:
        """Map the eodag product type to the provider product type"""
        if product_type is None:
            return None
        logger.debug("Mapping eodag product type to provider product type")
        return self.config.products.get(product_type, {}).get(
            "productType", GENERIC_PRODUCT_TYPE
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import unittest
from importlib.resources import files as res_files
from io import StringIO
from tempfile import TemporaryDirectory

import yaml.parser

from tests.context import (
    EXT_PRODUCT_TYPES_CONF_URI,
    HTTP_REQ_TIMEOUT,
    TEST_RESOURCES_PATH,
    USER_AGENT,
    EODataAccessGateway,
    ValidationError,
    config,
    get_ext_product_types_conf,
    load_stac_provider_config,
    merge_configs,
)
from tests.utils import mock


class TestProviderConfig(unittest.TestCase):
    def test_provider_config_name(self):
        """Name config parameter must be slugified"""
        unslugified_provider_name = "some $provider-name. Really ugly"
        slugified_provider_name = "some_provider_name_really_ugly"

        stream = StringIO(
            """!provider
            name: {}
            api: !plugin
                type: MyPluginClass
            products:
                EODAG_PRODUCT_TYPE: provider_product_type
            """.format(
                unslugified_provider_name
            )
        )
        provider_config = yaml.load(stream, Loader=yaml.Loader)
        self.assertEqual(provider_config.name, slugified_provider_name)

    def test_provider_config_valid(self):
        """Provider config must be valid"""
        # Not defining any plugin at all
        invalid_stream = StringIO("""!provider\nname: my_provider""")
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream, Loader=yaml.Loader
        )

        # Not defining a class for a plugin
        invalid_stream = StringIO(
            """!provider
                name: my_provider
                search: !plugin
                    param: value
            """
        )
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream, Loader=yaml.Loader
        )

        # Not giving a name to the provider
        invalid_stream = StringIO(
            """!provider
                api: !plugin
                    type: MyPluginClass
            """
        )
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream, Loader=yaml.Loader
        )

        # Specifying an api plugin and a search or download or auth plugin at the same
        # type
        invalid_stream1 = StringIO(
            """!provider
                api: !plugin
                    type: MyPluginClass
                search: !plugin
                    type: MyPluginClass2
            """
        )
        invalid_stream2 = StringIO(
            """!provider
                api: !plugin
                    type: MyPluginClass
                download: !plugin
                    type: MyPluginClass3
            """
        )
        invalid_stream3 = StringIO(
            """!provider
                api: !plugin
                    type: MyPluginClass
                auth: !plugin
                    type: MyPluginClass4
            """
        )
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream1, Loader=yaml.Loader
        )
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream2, Loader=yaml.Loader
        )
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream3, Loader=yaml.Loader
        )

    def test_provider_config_update(self):
        """A provider config must be update-able with a dict"""
        valid_stream = StringIO(
            """!provider
                name: provider
                provider_param: val
                api: !plugin
                    type: MyPluginClass
                    plugin_param1: value1
                    pluginParam2: value2
        """
        )
        provider_config = yaml.load(valid_stream, Loader=yaml.Loader)
        overrides = {
            "provider_param": "new val",
            "api": {"pluginparam2": "newVal", "newParam": "val"},
        }
        provider_config.update(overrides)
        self.assertEqual(provider_config.provider_param, "new val")
        self.assertEqual(provider_config.api.pluginParam2, "newVal")
        self.assertTrue(hasattr(provider_config.api, "newParam"))
        self.assertEqual(provider_config.api.newParam, "val")

    def test_provider_config_merge(self):
        """Merge 2 providers configs"""
        config_stream1 = StringIO(
            """!provider
                name: provider1
                provider_param: val
                provider_param2: val2
                api: !plugin
                    type: MyPluginClass
                    plugin_param1: value1
                    pluginParam2: value2
        """
        )
        config_stream2 = StringIO(
            """!provider
                name: provider1
                provider_param: val1
                provider_param3: val3
                api: !plugin
                    type: MyPluginClass
                    pluginParam2: value3
        """
        )
        provider_config1 = yaml.load(config_stream1, Loader=yaml.Loader)
        provider_config2 = yaml.load(config_stream2, Loader=yaml.Loader)

        providers_config = {
            "provider1": provider_config1,
            "provider2": provider_config1,
        }

        merge_configs(
            providers_config,
            {"provider1": provider_config2, "provider3": provider_config1},
        )
        self.assertEqual(len(providers_config), 3)
        self.assertEqual(providers_config["provider1"].provider_param, "val1")
        self.assertEqual(providers_config["provider1"].provider_param2, "val2")
        self.assertEqual(providers_config["provider1"].provider_param3, "val3")
        self.assertEqual(providers_config["provider1"].api.plugin_param1, "value1")
        self.assertEqual(providers_config["provider1"].api.pluginParam2, "value3")


class TestPluginConfig(unittest.TestCase):
    def test_plugin_config_valid(self):
        """A plugin config must specify a valid plugin type"""
        # A stream configuring a plugin without specifying the "type" key
        invalid_stream = StringIO(
            """!plugin
                    param: value
        """
        )
        self.assertRaises(
            ValidationError, yaml.load, invalid_stream, Loader=yaml.Loader
        )

        valid_stream = StringIO(
            """!plugin
                    type: MySearchPlugin
                    param1: value
        """
        )
        self.assertIsInstance(
            yaml.load(valid_stream, Loader=yaml.Loader), config.PluginConfig
        )

    def test_plugin_config_update(self):
        """A plugin config must be update-able by a dict"""
        valid_stream = StringIO(
            """!plugin
                    type: MyPluginClass
                    plugin_param1: value1
                    pluginParam2:
                        sub_param1: v1
                        subParam_2: v2
        """
        )
        plugin_config = yaml.load(valid_stream, Loader=yaml.Loader)
        overrides = {
            "type": "MyOtherPlugin",
            "new_plugin_param": "a value",
            "pluginparam2": {"sub_param1": "new_val1"},
        }
        plugin_config.update(overrides)
        self.assertEqual(plugin_config.type, "MyOtherPlugin")
        self.assertEqual(plugin_config.pluginParam2["sub_param1"], "new_val1")
        self.assertTrue(hasattr(plugin_config, "new_plugin_param"))
        self.assertEqual(plugin_config.new_plugin_param, "a value")


class TestConfigFunctions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestConfigFunctions, cls).setUpClass()
        # mock os.environ to empty env
        cls.mock_os_environ = mock.patch.dict(os.environ, {}, clear=True)
        cls.mock_os_environ.start()

    @classmethod
    def tearDownClass(cls):
        super(TestConfigFunctions, cls).tearDownClass()
        # stop os.environ
        cls.mock_os_environ.stop()

    def test_load_default_config(self):
        """Default config must be successfully loaded"""
        conf = config.load_default_config()
        self.assertIsInstance(conf, dict)
        for key, value in conf.items():
            # keys of the default conf dict are the names of the provider
            self.assertEqual(key, value.name)
            # providers implementing download or api store their downloaded products in
            # tempdir by default
            download_plugin = getattr(value, "download", getattr(value, "api", None))
            if download_plugin is not None:
                self.assertEqual(download_plugin.output_dir, tempfile.gettempdir())
            # priority is set to 0 unless you are 'peps' provider
            if key == "peps":
                self.assertEqual(value.priority, 1)
            else:
                self.assertEqual(value.priority, 0)

    def test_load_config_providers_whitelist(self):
        """Config must be loaded with only the selected whitelist of providers"""
        try:
            os.environ["EODAG_PROVIDERS_WHITELIST"] = "creodias"
            conf = config.load_default_config()

            self.assertIsInstance(conf, dict)
            self.assertEqual({"creodias"}, set(conf.keys()))
        finally:
            os.environ.pop("EODAG_PROVIDERS_WHITELIST", None)

    def test_override_config_from_str(self):
        """Default configuration must be overridden from a yaml conf str"""
        default_config = config.load_default_config()
        conf_update = yaml.safe_load(
            """
            my_new_provider:
                priority: 4
                search:
                    type: StacSearch
                    api_endpoint: https://api.my_new_provider/search
                products:
                    S2_MSI_L1C:
                        productType: sentinel2_l1c
                    GENERIC_PRODUCT_TYPE:
                        productType: '{productType}'
                download:
                    type: AwsDownload
                    s3_endpoint: https://api.my_new_provider
                auth:
                    type: AwsAuth
                    credentials:
                        aws_access_key_id: access-key-id
                        aws_secret_access_key: secret-access-key
            """
        )
        config.override_config_from_mapping(default_config, conf_update)

        my_new_provider_conf = default_config["my_new_provider"]
        self.assertEqual(my_new_provider_conf.priority, 4)
        self.assertIsInstance(my_new_provider_conf.search, config.PluginConfig)
        self.assertEqual(
            my_new_provider_conf.products["S2_MSI_L1C"]["productType"], "sentinel2_l1c"
        )
        self.assertEqual(
            my_new_provider_conf.auth.credentials["aws_secret_access_key"],
            "secret-access-key",
        )

    def test_override_config_from_file(self):
        """Default configuration must be overridden from a conf file

        # noqa: E800
        Content of file_config_override.yml
        usgs:
          priority: 5
          api:
              extract: False
              credentials:
                  username: usr
                  password: pwd

        aws_eos:
          search:
              product_location_scheme: file
          search_auth:
              credentials:
                  apikey: api-key
          download_auth:
              credentials:
                  aws_access_key_id: access-key-id
                  aws_secret_access_key: secret-access-key

        peps:
          download:
              output_dir: /data

        my_new_provider:
            priority: 4
            search:
                type: StacSearch
                api_endpoint: https://api.my_new_provider/search
            products:
                S2_MSI_L1C:
                  productType: sentinel2_l1c
                GENERIC_PRODUCT_TYPE:
                  productType: '{productType}'
            download:
                type: AwsDownload
                s3_endpoint: https://api.my_new_provider
                flatten_top_dirs: false
            auth:
                type: AwsAuth
                credentials:
                  aws_access_key_id: access-key-id
                  aws_secret_access_key: secret-access-key
        """
        default_config = config.load_default_config()
        file_path_override = os.path.join(
            os.path.dirname(__file__), "resources", "file_config_override.yml"
        )

        config.override_config_from_file(default_config, file_path_override)
        usgs_conf = default_config["usgs"]
        self.assertEqual(usgs_conf.priority, 5)
        self.assertEqual(usgs_conf.api.extract, False)
        self.assertEqual(usgs_conf.api.credentials["username"], "usr")
        self.assertEqual(usgs_conf.api.credentials["password"], "pwd")

        aws_conf = default_config["aws_eos"]
        self.assertEqual(aws_conf.search.product_location_scheme, "file")
        self.assertEqual(aws_conf.search_auth.credentials["apikey"], "api-key")
        self.assertEqual(
            aws_conf.download_auth.credentials["aws_access_key_id"], "access-key-id"
        )
        self.assertEqual(
            aws_conf.download_auth.credentials["aws_secret_access_key"],
            "secret-access-key",
        )

        peps_conf = default_config["peps"]
        self.assertEqual(peps_conf.download.output_dir, "/data")

        my_new_provider_conf = default_config["my_new_provider"]
        self.assertEqual(my_new_provider_conf.priority, 4)
        self.assertIsInstance(my_new_provider_conf.search, config.PluginConfig)
        self.assertEqual(my_new_provider_conf.search.type, "StacSearch")
        self.assertEqual(
            my_new_provider_conf.search.api_endpoint,
            "https://api.my_new_provider/search",
        )
        self.assertIsInstance(my_new_provider_conf.products, dict)
        self.assertEqual(
            my_new_provider_conf.products["S2_MSI_L1C"]["productType"], "sentinel2_l1c"
        )
        self.assertEqual(
            my_new_provider_conf.products["GENERIC_PRODUCT_TYPE"]["productType"],
            "{productType}",
        )
        self.assertIsInstance(my_new_provider_conf.download, config.PluginConfig)
        self.assertEqual(my_new_provider_conf.download.type, "AwsDownload")
        self.assertEqual(
            my_new_provider_conf.download.s3_endpoint, "https://api.my_new_provider"
        )
        self.assertFalse(my_new_provider_conf.download.flatten_top_dirs)
        self.assertIsInstance(my_new_provider_conf.auth, config.PluginConfig)
        self.assertEqual(my_new_provider_conf.auth.type, "AwsAuth")
        self.assertEqual(
            my_new_provider_conf.auth.credentials["aws_access_key_id"], "access-key-id"
        )
        self.assertEqual(
            my_new_provider_conf.auth.credentials["aws_secret_access_key"],
            "secret-access-key",
        )

    def test_override_config_from_env(self):
        """Default configuration must be overridden by environment variables"""
        default_config = config.load_default_config()
        os.environ["EODAG__USGS__PRIORITY"] = "5"
        os.environ["EODAG__USGS__API__EXTRACT"] = "false"
        os.environ["EODAG__USGS__API__CREDENTIALS__USERNAME"] = "usr"
        os.environ["EODAG__USGS__API__CREDENTIALS__PASSWORD"] = "pwd"
        os.environ["EODAG__AWS_EOS__SEARCH__PRODUCT_LOCATION_SCHEME"] = "file"
        os.environ["EODAG__AWS_EOS__SEARCH_AUTH__CREDENTIALS__APIKEY"] = "api-key"
        os.environ[
            "EODAG__AWS_EOS__DOWNLOAD_AUTH__CREDENTIALS__AWS_ACCESS_KEY_ID"
        ] = "access-key-id"
        os.environ[
            "EODAG__AWS_EOS__DOWNLOAD_AUTH__CREDENTIALS__AWS_SECRET_ACCESS_KEY"
        ] = "secret-access-key"
        os.environ["EODAG__PEPS__DOWNLOAD__OUTPUT_DIR"] = "/data"
        # check a parameter that has not been set yet
        self.assertFalse(hasattr(default_config["peps"].search, "timeout"))
        self.assertNotIn("start_page", default_config["peps"].search.pagination)
        os.environ["EODAG__PEPS__SEARCH__TIMEOUT"] = "3.1"
        os.environ["EODAG__PEPS__SEARCH__PAGINATION__START_PAGE"] = "2"

        config.override_config_from_env(default_config)
        usgs_conf = default_config["usgs"]
        self.assertEqual(usgs_conf.priority, 5)
        self.assertEqual(usgs_conf.api.extract, False)
        self.assertEqual(usgs_conf.api.credentials["username"], "usr")
        self.assertEqual(usgs_conf.api.credentials["password"], "pwd")

        aws_conf = default_config["aws_eos"]
        self.assertEqual(aws_conf.search.product_location_scheme, "file")
        self.assertEqual(aws_conf.search_auth.credentials["apikey"], "api-key")
        self.assertEqual(
            aws_conf.download_auth.credentials["aws_access_key_id"], "access-key-id"
        )
        self.assertEqual(
            aws_conf.download_auth.credentials["aws_secret_access_key"],
            "secret-access-key",
        )

        peps_conf = default_config["peps"]
        self.assertEqual(peps_conf.download.output_dir, "/data")
        self.assertEqual(peps_conf.search.timeout, 3.1)
        self.assertEqual(peps_conf.search.pagination["start_page"], 2)

    @mock.patch("requests.get", autospec=True)
    def test_get_ext_product_types_conf(self, mock_get):
        """External product types configuration must be loadable from remote or local file"""
        ext_product_types_path = os.path.join(
            TEST_RESOURCES_PATH, "ext_product_types.json"
        )

        # mock get request response for remote conf file (default value)
        mock_get.return_value = mock.Mock()
        mock_get.return_value.json.return_value = {"some_parameter": "a_value"}

        ext_product_types_conf = get_ext_product_types_conf()
        mock_get.assert_called_once_with(
            EXT_PRODUCT_TYPES_CONF_URI, headers=USER_AGENT, timeout=HTTP_REQ_TIMEOUT
        )
        self.assertEqual(ext_product_types_conf, {"some_parameter": "a_value"})

        # local conf file
        ext_product_types_conf = get_ext_product_types_conf(ext_product_types_path)
        self.assertIsInstance(ext_product_types_conf, dict)
        self.assertIn("foo", ext_product_types_conf["earth_search"]["providers_config"])


class TestStacProviderConfig(unittest.TestCase):
    def setUp(self):
        super(TestStacProviderConfig, self).setUp()
        # Mock home and eodag conf directory to tmp dir
        self.tmp_home_dir = TemporaryDirectory()
        self.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, return_value=self.tmp_home_dir.name
        )
        self.expanduser_mock.start()

        self.dag = EODataAccessGateway()

    def tearDown(self):
        super(TestStacProviderConfig, self).tearDown()
        # stop Mock and remove tmp config dir
        self.expanduser_mock.stop()
        self.tmp_home_dir.cleanup()

    def test_existing_stac_provider_conf(self):
        """Existing / pre-configured STAC providers conf should mix providers.yml and  stac_provider.yml infos."""
        with open(str(res_files("eodag") / "resources" / "providers.yml"), "r") as fh:
            providers_configs = {
                p.name: p for p in yaml.load_all(fh, Loader=yaml.Loader)
            }

        raw_provider_search_conf = providers_configs["usgs_satapi_aws"].search.__dict__
        common_stac_provider_search_conf = load_stac_provider_config()["search"]
        provider_search_conf = self.dag.providers_config[
            "usgs_satapi_aws"
        ].search.__dict__

        # conf existing in common (stac_provider.yml) and not in raw_provider (providers.yml)
        self.assertIn(
            "resolution", common_stac_provider_search_conf["metadata_mapping"]
        )
        self.assertNotIn("resolution", raw_provider_search_conf["metadata_mapping"])
        self.assertIn("resolution", provider_search_conf["metadata_mapping"])

        self.assertIn("discover_metadata", common_stac_provider_search_conf)
        self.assertNotIn("discover_metadata", raw_provider_search_conf)
        self.assertIn("discover_metadata", provider_search_conf)

        # raw_provider conf (providers.yml) should overwrite common conf (stac_provider.yml)
        self.assertEqual(
            raw_provider_search_conf["metadata_mapping"]["assets"],
            provider_search_conf["metadata_mapping"]["assets"],
        )
        self.assertNotEqual(
            common_stac_provider_search_conf["metadata_mapping"]["assets"],
            provider_search_conf["metadata_mapping"]["assets"],
        )

        # check if raw_provider_search_conf is a subset of provider_search_conf
        for k, v in raw_provider_search_conf.items():
            if isinstance(v, dict):
                assert (
                    raw_provider_search_conf[k].items()
                    <= provider_search_conf[k].items()
                )
            else:
                self.assertEqual(v, provider_search_conf[k])

    def test_custom_stac_provider_conf(self):
        """Custom STAC providers conf should mix providers.yml and stac_provider.yml infos."""
        custom_stac_provider_conf_yml = """
            foo:
                search:
                    type: StacSearch
                    api_endpoint: https://foo.bar/search
                    metadata_mapping:
                        title: '$.properties."foo:bar_baz"'
                products:
                    GENERIC_PRODUCT_TYPE:
                        productType: '{productType}'
                download:
                    type: HTTPDownload
                    base_uri: https://foo.bar
        """
        self.dag.update_providers_config(custom_stac_provider_conf_yml)
        custom_stac_provider_conf = yaml.safe_load(custom_stac_provider_conf_yml)[
            "foo"
        ]["search"]

        common_stac_provider_search_conf = load_stac_provider_config()["search"]
        provider_search_conf = self.dag.providers_config["foo"].search.__dict__

        # conf existing in common (stac_provider.yml) and not in raw_provider (providers.yml)
        self.assertIn(
            "resolution", common_stac_provider_search_conf["metadata_mapping"]
        )
        self.assertNotIn("resolution", custom_stac_provider_conf["metadata_mapping"])
        self.assertIn("resolution", provider_search_conf["metadata_mapping"])

        self.assertIn("discover_metadata", common_stac_provider_search_conf)
        self.assertNotIn("discover_metadata", custom_stac_provider_conf)
        self.assertIn("discover_metadata", provider_search_conf)

        # raw_provider conf (providers.yml) should overwrite common conf (stac_provider.yml)
        self.assertEqual(
            custom_stac_provider_conf["metadata_mapping"]["title"],
            provider_search_conf["metadata_mapping"]["title"],
        )
        self.assertNotEqual(
            common_stac_provider_search_conf["metadata_mapping"]["title"],
            provider_search_conf["metadata_mapping"]["title"],
        )

        # check if custom_stac_provider_conf is a subset of provider_search_conf
        for k, v in custom_stac_provider_conf.items():
            if isinstance(v, dict):
                assert (
                    custom_stac_provider_conf[k].items()
                    <= provider_search_conf[k].items()
                )
            else:
                self.assertEqual(v, provider_search_conf[k])
